
#include "netscan.h"

extern struct in_addr ip_src;

extern int debug;

static unsigned short ic_check(void* , int );

extern void fin(int);

extern int read_ip_packet(char *dst, int len);

void icmp_ping(struct in_addr ddir, unsigned short id, int fd)
{
  char buffer[1024];
  struct icmp *icmph = (struct icmp*) buffer;
  struct sockaddr_in to;

  memset(buffer, 0, 1024);

  icmph->icmp_type = 8;
  icmph->icmp_hun.ih_idseq.icd_id = id;
  icmph->icmp_cksum = ic_check(icmph, sizeof(struct icmp));

  to.sin_family = AF_INET;
  to.sin_addr.s_addr = ddir.s_addr;

  if (sendto(fd, buffer, sizeof(struct icmp), 0, &to, sizeof(to)) < 0) {
    printf(" error: icmp_ping(): sendto() devolvio -1\nsendto(): %s\n", strerror(errno));
    fin(0);
  }
}

struct in_addr icmp_wait_pong(unsigned short id, int timeout)
{
  struct in_addr ret;
  char buffer[1024];
  struct ip *iph;
  struct icmp *icmph;

  ret.s_addr = 0;

  for (;;) {

    if (!read_ip_packet(buffer, 20)) {
      timeout--;
      if (!timeout) return ret;
      usleep(1);
      continue;
    }

    iph = (struct ip*) buffer;

    if (iph->ip_p != 1) continue;
    
    if (!ip_src.s_addr) 
      ip_src.s_addr = iph->ip_dst.s_addr;
    
    icmph = (struct icmp*) (buffer + (iph->ip_hl * 4));

    if (icmph->icmp_type != 0) continue;
    
    if (icmph->icmp_hun.ih_idseq.icd_id != id) continue;
    
    ret.s_addr = iph->ip_src.s_addr;
    return ret;
  }
}

static unsigned short ic_check(void *dataptr, int len)
{
  unsigned long acum = 0;
  unsigned short ret,*ptr;
  char *pad;
  int c;

  ptr = (unsigned short*) dataptr;
  for (c = 0; c < len; c+=2) {
    acum += (*(ptr++));
  }
  if (c < len) {
    ptr--;
    pad = (char*) ptr;
    pad++;
    acum += ((*pad) << 8);
  }

  ret = ~(((acum >> 16) + (acum & 0xFFFF)) & 0xFFFF); 
  return ret;
}

void icmp_tstamp(struct in_addr ddir, unsigned short id, int fd)
{
  char buffer[1024];
  struct icmp *icmph = (struct icmp*) buffer;
  struct sockaddr_in to;
  struct timeval tv;
  struct timezone tz;

  memset(buffer, 0, 1024);

  memset(&tz, 0, sizeof(tz));
  gettimeofday(&tv, &tz);

  icmph->icmp_type = 13;
  icmph->icmp_hun.ih_idseq.icd_id = id;
  icmph->icmp_hun.ih_idseq.icd_seq = 0;
  icmph->icmp_dun.id_ts.its_otime = htonl((tv.tv_sec % 86400) * 1000 + tv.tv_usec / 1000);
  icmph->icmp_cksum = ic_check(icmph, sizeof(struct icmp));

  to.sin_family = AF_INET;
  to.sin_addr.s_addr = ddir.s_addr;

  if (sendto(fd, buffer, sizeof(struct icmp), 0, &to, sizeof(to)) < 0) {
    printf(" error: icmp_tstamp(): sendto() devolvio -1\nsendto(): %s\n", strerror(errno));
    exit(0);
  }
}

struct in_addr icmp_wait_tstamp(unsigned short id, int timeout, u_int32_t *dif)
{
  struct in_addr ret;
  char buffer[1024];
  struct ip *iph;
  struct icmp *icmph;
  u_int32_t tmp;

  ret.s_addr = 0;

  for (;;) {

    if (!read_ip_packet(buffer, 20)) {
      timeout--;
      if (!timeout) return ret;
      usleep(1);
      continue;
    }

    iph = (struct ip*) buffer;

    if (iph->ip_p != 1) continue;

    if (!ip_src.s_addr) 
      ip_src.s_addr = iph->ip_dst.s_addr;

    icmph = (struct icmp*) (buffer + (iph->ip_hl * 4));

    if (icmph->icmp_type != 14) continue;
    
    if (icmph->icmp_hun.ih_idseq.icd_id != id) continue;
    
    ret.s_addr = iph->ip_src.s_addr;
    
    tmp = htonl(icmph->icmp_dun.id_ts.its_ttime) - htonl(icmph->icmp_dun.id_ts.its_otime);

    tmp /= 1000;

    tmp >>= 2;

    *dif = tmp;
    return ret;
  }
}
