
#include "netscan.h"

extern struct in_addr mi_ip();
extern void IP_TCP_OPT(int, struct in_addr, struct in_addr, u_int8_t, u_int32_t, u_int16_t, u_int16_t, u_int16_t, int, char*);
extern int misocket(int, int, int);
extern void set_timeout();
extern int get_timeout();
extern int read_ip_packet(char* , int);
extern int exists(char* );
extern int verbose;
extern FILE *logfile;

int guess = 0;

struct respuesta {
  u_int8_t resp;
  u_int8_t dont_frag;
  u_int8_t ack;
  u_int16_t flags;
  u_int16_t window;
  int nopts;
  char *opts;
};

struct huella {
  char SO[256];
  struct respuesta r1; /* FIN a un puerto abierto */
  struct respuesta r2; /* ACK|URG|PSH|FIN con opciones a un puerto abierto */
  struct respuesta r3; /* SYN|ACK con opciones a un puerto cerrado */
  struct respuesta r4; /* SYN con opciones a un puerto abierto */
  struct respuesta r5; /* ACK con opciones a un puerto cerrado */
  struct respuesta r6; /* SYN + flag indefinido con opciones a un puerto abierto */
  struct respuesta r7; /* NULL a un puerto abierto */
};

char *SOPTS = 
  "\x02\x04\x00\xff" /* MSS */
  "\x03\x03\x0a"     /* window scaling */
  "\x08\x0a\x01\x01\x01\x01\x01\x01\x01\x01" /* Time Stamp */
  "\x01\x01\x00"; /* NOP NOP EOL */

char OPTS[1024];

int cmp_resp(struct respuesta *r1, struct respuesta *r2)
{
  if (r1->resp == r2->resp)
    if (r1->dont_frag == r2->dont_frag)
      if (r1->ack == r2->ack)
	if (r1->flags == r2->flags)
	  if (r1->window == r2->window)
	    if (r1->nopts == r2->nopts) return 0;
	      //	      if (!strcmp(r1->opts, r2->opts)) return 0;
  return 1;
}

int get_nopts(struct tcphdr *tcph)
{
  char *opts = (char*) ((char*)tcph + 20);
  int len = (tcph->th_off * 4) - 20;
  int c;
  int ret = 0;

  memset(OPTS, 0, 1024);

  for (c = 0; c < len; c++) {
    
    if (opts[c] == 0) { 
      OPTS[ret++] = 'E';
      continue;
    }

    if (opts[c] == 1) { 
      OPTS[ret++] = 'N';
      continue;
    }

    if (opts[c] == 2) { 
      OPTS[ret++] = 'M';
      c += 3;
      continue;
    }

    if (opts[c] == 3) { 
      OPTS[ret++] = 'W';
      c += 2;
      continue;
    }

    if (opts[c] == 4) { 
      OPTS[ret++] = 'S';
      c++;
      continue;
    }
    
    if (opts[c] == 5) { 
      OPTS[ret++] = 's';
      continue;
    }
    
    if (opts[c] == 8) { 
      OPTS[ret++] = 'T';
      c += 9;
      continue;
    }
  }
  return ret;
}

void test(struct in_addr dir, struct respuesta *dest, u_int16_t port, u_int8_t flags, int optl, char *opts)
{
  int fd = misocket(AF_INET, SOCK_RAW, 255);
  char buff[4096];
  struct ip *iph = (struct ip*) buff;
  struct tcphdr *tcph = NULL;
  int psrc = rand() % 0xffff;
  int resp = 0;
  u_int8_t ttl = rand() % 0xff;
  u_int32_t nseq = rand();

  IP_TCP_OPT(fd, mi_ip(), dir, ttl, nseq, psrc, port, flags, optl, opts);
  
  set_timeout();
  
  while (1) {
    
    memset(buff, 0, 4096);
    
    if (read_ip_packet(buff, 4096 - 20))
      if (iph->ip_p == 6)
	if (iph->ip_src.s_addr == dir.s_addr) {
	  tcph = (struct tcphdr*) (buff + (iph->ip_hl * 4));
	  if ((tcph->th_dport == htons(psrc)) &&
	      (tcph->th_sport == htons(port))) {
	    resp = 1;
	    break;
	  }
	}
    if (get_timeout()) break;
  }
		     
  if (!resp) return;

  dest->resp = 1;
  if (htons(iph->ip_off) & IP_DF) dest->dont_frag = 1;
  if (tcph->th_ack == nseq) dest->ack = 1;
  if (htonl(tcph->th_ack) == (htonl(nseq) + 1)) dest->ack = 2;
  dest->flags = tcph->th_flags;
  dest->window = htons(tcph->th_win);
  dest->nopts = get_nopts(tcph);
  if (dest->nopts) {
    dest->opts = (char*) malloc(dest->nopts + 1);
    strcpy(dest->opts, OPTS);
  }
}


struct huella *get_huella_host(struct in_addr dir, int oport, int cport)
{
  struct huella *ret = NULL;
  int c;

  ret = (struct huella*) malloc(sizeof(struct huella));
  memset(ret, 0, sizeof(struct huella));

  for (c = 0; c < 3; c++) {
    if (!ret->r1.resp) test(dir, &ret->r1, oport, TH_FIN, 0, NULL);
    if (!ret->r2.resp) test(dir, &ret->r2, oport, TH_ACK | TH_URG | TH_FIN | TH_PUSH, 20, SOPTS);
    if (!ret->r3.resp) test(dir, &ret->r3, cport, TH_SYN | TH_ACK, 20, SOPTS);
    if (!ret->r4.resp) test(dir, &ret->r4, oport, TH_SYN, 20, SOPTS);
    if (!ret->r5.resp) test(dir, &ret->r5, cport, TH_ACK, 20, SOPTS);
    if (!ret->r6.resp) test(dir, &ret->r6, oport, TH_SYN | 0x80, 20, SOPTS);
    if (!ret->r7.resp) test(dir, &ret->r7, cport, 0, 0, NULL);
  }
  return ret;
}

void print_huella(struct huella *h)
{
  struct respuesta *r = NULL;
  int c;

  for (c = 0; c < 7; c++) {
    switch(c) {
    case 0:
      r = &h->r1; break;
    case 1:
      r = &h->r2; break;
    case 2:
      r = &h->r3; break;
    case 3:
      r = &h->r4; break;
    case 4:
      r = &h->r5; break;
    case 5:
      r = &h->r6; break;
    case 6:
      r = &h->r7; break;
    }
    printf("R%i(", c + 1);
    printf("RESP=%x:DF=%x:NACK=%x:FLAGS=%x:WIN=%x:NOPTS=%x:OPTS=%s)\n", r->resp, r->dont_frag, r->ack, 
	   r->flags, r->window, r->nopts, (r->opts ? r->opts : "NO_OPTS"));
  }
}

void read_comment(FILE *f)
{
  char ch;
  do {
    if (feof(f)) return;
    fread(&ch, 1, 1, f);
  } while (ch != '\n');
}

char read_space(FILE *f)
{
  char ch;
  do {
    if (feof(f)) return 0;
    fread(&ch, 1, 1, f);
  } while ((ch == '\n') || (ch == ' '));
  return ch;
}

void compara_huella(struct huella *h)
{
  FILE *fh = fopen("huellas.dat", "r");
  struct huella tmp;
  struct respuesta *r = NULL, *s = NULL;
  char ch;
  int c, d;
  char buf[256];
  char *ptr;

  if (!fh) {
    printf("Error abriendo huellas.dat\n");
    perror("open()");
    exit(0);
  }

  while (!feof(fh)) {

    for (;;) {
      ch = read_space(fh);
      if (ch == '#') read_comment(fh);
      else break;
      if (feof(fh)) break;
    }
    
    if (feof(fh)) break;
    
    c = 0;
    memset(&tmp, 0, sizeof(tmp));
    tmp.SO[c++] = ch;

    for (;;) {
      
      if (feof(fh)) break;
      fread(&ch, 1, 1, fh);
      if ((ch == '#') || (ch == '\n')) break; 
      tmp.SO[c++] = ch;
    }
    if (feof(fh)) break;

    if (ch == '#') read_comment(fh);

    for (c = 0; c < 7; c++) {
      switch(c) {
      case 0:
	r = &tmp.r1; break;
      case 1:
	r = &tmp.r2; break;
      case 2:
	r = &tmp.r3; break;
      case 3:
	r = &tmp.r4; break;
      case 4:
	r = &tmp.r5; break;
      case 5:
	r = &tmp.r6; break;
      case 6:
	r = &tmp.r7; break;
      }

      if (feof(fh)) break;

      fgets(buf, 256, fh); 

      if (!r->opts) r->opts = (char*) malloc(256);
      memset(r->opts, 0, 256);

      sscanf(buf, "R%i(RESP=%x:DF=%x:NACK=%x:FLAGS=%x:WIN=%x:NOPTS=%x:OPTS=)", &d, (int*)&r->resp, (int*)&r->dont_frag, (int*)&r->ack, 
      	     (int*)&r->flags, (int*)&r->window, (int*)&r->nopts);

      ptr = strstr(buf, "OPTS=");

      if ((d != (c+1)) || (!ptr))  {
	printf("Error: El fichero huellas.dat tiene un formato no valido\n");
	exit(0);
      }

      ptr += 5;
      for (d = 0; d < 256; d++) {
	if (ptr[d] == ')') break;
	r->opts[d] = ptr[d];
      }

      if (feof(fh)) break;
    }
    
    for (c = 0; c < 7; c++) {
      switch(c) {
      case 0:
	r = &tmp.r1; 
	s = &h->r1; break;
      case 1:
	s = &h->r2;
	r = &tmp.r2; break;
      case 2:
	s = &h->r3;
	r = &tmp.r3; break;
      case 3:
	s = &h->r4;
	r = &tmp.r4; break;
      case 4:
	s = &h->r5;
	r = &tmp.r5; break;
      case 5:
	s = &h->r6;
	r = &tmp.r6; break;
      case 6:
	s = &h->r7;
	r = &tmp.r7; break;
      }
      d = 0;
      if (cmp_resp(r, s)) break;
      d = 1;
    }

    if ((c == 7) && (d == 1)) {
      printf("SO: %s\n", tmp.SO);
      if (logfile) fprintf(logfile, "SO: %s\n", tmp.SO);
      guess = 1;
      goto libera;
    }

    if (feof(fh)) break;
  }

  printf("SO: Desconocido\n");
  if (logfile) fprintf(logfile, "SO: Desconocido\n");
  print_huella(h);

 libera:
  
  fclose(fh);

  for (c = 0; c < 7; c++) {
    switch(c) {
    case 0:
      r = &tmp.r1; break;
    case 1:
      r = &tmp.r2; break;
    case 2:
      r = &tmp.r3; break;
    case 3:
      r = &tmp.r4; break;
    case 4:
      r = &tmp.r5; break;
    case 5:
      r = &tmp.r6; break;
    case 6:
      r = &tmp.r7; break;
    }
    if (r->opts) free(r->opts);
  }
}

void osdetect(struct in_addr dir, int oport, int cport)
{
  struct huella *h;
  if (!exists("huellas.dat")) {
    printf("Error: no encuentro el fichero huellas.dat. Saltando este escaneo\n");
    return;
  }

  h = get_huella_host(dir, oport, cport);

  guess = 0;
  compara_huella(h);
  
  if (guess)
    if (verbose > 1) print_huella(h);
  free(h);
}
