
#include "netscan.h"

extern struct host *lista_hosts;

extern int ack_port, nhosts, solohosts, npuertos, noresuelve, solo_pings;

extern FILE *logfile;

extern void fin(int);

struct host *tmp_hosts = NULL;

int ntmphosts = 0;

extern int scan_timeout_sec, scan_timeout_msec;

extern struct in_addr icmp_wait_pong(unsigned short id, int timeout);
extern struct in_addr TCP_wait_ack(unsigned short sp, int timeout);
extern struct in_addr mi_ip();
extern void icmp_ping(struct in_addr ddir, unsigned short id, int fd);   
extern void TCP(int fd, struct in_addr sdir, struct in_addr ddir, 
	 unsigned short sp, unsigned short dp, unsigned char flags);
extern void set_timeout();
extern int get_timeout();
extern int misocket(int, int, int);

extern int MAX_PACKETS;

int seq = 1;

void resuelve_a_char(struct in_addr dir, char *dst, int len)
{
  struct hostent *he;

  he = gethostbyaddr((char*)&dir, sizeof(dir), AF_INET);

  if (!he) return;
  strncpy(dst, he->h_name, len);
}

u_int32_t resuelve_a_ip(char *host)
{
  struct hostent *he = NULL;
  u_int32_t ret;

  if ((ret = inet_addr(host))!= -1) return ret;

  he = gethostbyname(host);
  if (!he) {
    printf(" Error: gethostbyname(): %s no se puede resolver\n", host);
    exit(0);
  }
  
  strncpy((char*)&ret, (char*)he->h_addr, 4);
  return ret;
}

int busca_host(struct in_addr dir)
{
  struct host *ptr = lista_hosts;
  while (ptr) {
    if (ptr->dir.s_addr == dir.s_addr) return 1;
    ptr = ptr->next;
  }
  return 0;
}

int busca_tmp_host(struct in_addr dir)
{
  struct host *ptr = tmp_hosts;
  while (ptr) {
    if (ptr->dir.s_addr == dir.s_addr) return 1;
    ptr = ptr->next;
  }
  return 0;
}

struct host *get_tmp_host(struct in_addr dir)
{
  struct host *ptr = tmp_hosts;
  while (ptr) {
    if (ptr->dir.s_addr == dir.s_addr) return ptr;
    ptr = ptr->next;
  }
  return NULL;
}

void new_tmp_host(struct in_addr dir)
{
  struct host *ptr = tmp_hosts, *new;

  if (busca_tmp_host(dir)) return;

  if (!ptr) {
    tmp_hosts = (struct host*) malloc(sizeof(struct host));
    memset(tmp_hosts, 0, sizeof(struct host));
    new = tmp_hosts;
    new->dir.s_addr = dir.s_addr;
    new->seq = seq++;
    ntmphosts = 1;
    return;
  }
  while (ptr->next) ptr = ptr->next;
  new = (struct host*) malloc(sizeof(struct host));
  memset(new, 0, sizeof(struct host));
  new->dir.s_addr = dir.s_addr;
  new->seq = seq++;
  ptr->next = new;
  ntmphosts++;
}

struct hash {
  struct host *addr;
};

void host_timeout(struct host *ht)
{ 
  gettimeofday(&ht->time1, NULL);
}

void calcula_host_timeout(struct host *ht)
{
  gettimeofday(&ht->time2, NULL);
  ht->timeout.tv_sec = ht->time2.tv_sec - ht->time1.tv_sec; 
  ht->timeout.tv_usec = ht->time2.tv_usec - ht->time1.tv_usec; 
}

void resuelve_host()
{
  struct host *ptr, *new = lista_hosts;
  struct hash *hashlist;
  int c, d = 0;
  int seq_zero = nhosts - 1; 

  hashlist = (struct hash*) malloc(sizeof(struct hash) * nhosts);
  for (c = 0; c < sizeof(struct hash); c++) hashlist[c].addr = 0;
  
  for (c = 1, d = 0; c < (ntmphosts + 1); c++, d++) {
    ptr = lista_hosts;
    while (ptr) {
      if (ptr->seq == c)  break;
      ptr = ptr->next;
    }
    if (ptr) hashlist[d].addr = ptr;
    else d--;
  }

  ptr = lista_hosts;

  for (c = 0; c < nhosts; c++) {
    
    if (ptr->seq) {
      ptr = ptr->next;
      continue;
    }

    hashlist[seq_zero--].addr = ptr;
    ptr = ptr->next;
  }

  lista_hosts = hashlist[0].addr;
  ptr = lista_hosts;

  for (c = 1; c < nhosts; c++) {
    ptr->next = hashlist[c].addr;
    if (c == (nhosts - 1)) {
      ptr = ptr->next;
      ptr->next = NULL;
    }    
    else ptr = ptr->next;
  }
  
  new = lista_hosts;
  c = 0;

  if (!noresuelve) {
    if (verbose) printf("Resolviendo %i hosts\n", nhosts);
    fflush(stdout);

    while (new) {
      resuelve_a_char(new->dir, new->name, 1024);
      new = new->next;
    }
  }
 
}

void new_host(struct in_addr dir)
{
  struct host *tmp, *ptr = lista_hosts, *new;

  if (busca_host(dir)) return;
  
  tmp = get_tmp_host(dir);

  if (!ptr) {
    lista_hosts = (struct host*) malloc(sizeof(struct host));
    memset(lista_hosts, 0, sizeof(struct host));
    new = lista_hosts;
    new->dir.s_addr = dir.s_addr;
    new->puertos = (struct puerto*) malloc(npuertos * sizeof(struct puerto));
    
    if (tmp) {
      new->seq = tmp->seq;
      memcpy(&new->time1, &tmp->time1, sizeof(tmp->time1));
      calcula_host_timeout(new);
    } else {
      new->timeout.tv_sec = scan_timeout_sec;
      new->timeout.tv_usec = scan_timeout_msec;
    }
    nhosts = 1;
    return;
  }
 
  while (ptr->next) ptr = ptr->next;
  new = (struct host*) malloc(sizeof(struct host));
  memset(new, 0, sizeof(struct host));
  new->dir.s_addr = dir.s_addr;
  new->puertos = (struct puerto*) malloc(npuertos * sizeof(struct puerto));
  ptr->next = new;
  if (tmp) {
    new->seq = tmp->seq;
    memcpy(&new->time1, &tmp->time1, sizeof(tmp->time1));
    calcula_host_timeout(new);
  } else {
    new->timeout.tv_sec = scan_timeout_sec;
    new->timeout.tv_usec = scan_timeout_msec;
  }
  nhosts++;
}

void parse_mask(struct in_addr dir, int cmask)
{
  int mask = 0;
  int loops = 0, c;
  u_int32_t d = dir.s_addr;

  if ((cmask < 1) || (cmask > 30)) {
    printf(" Error: mascara invalida\n");
    fin(0);
  }

  c = cmask;
  while (cmask--) mask = (mask >> 1) | 0x80000000;
  cmask = 32 - c;
  while (cmask--) loops = (loops << 1) | 0x1;

  for (c = 1; c <= loops; c++) {
    d = htonl(d);
    d = (d & mask) | c;
    d = htonl(d);
    dir.s_addr = d;
    new_tmp_host(dir);
  }
}

void parse_target(char *target)
{
  char tmp[1024], *pmask;
  int c, mask;
  struct in_addr dir;

  if (strstr(target, "/")) {
    if (strstr(target, "*")) {
      printf("Error: no puedes usar / y * a la vez en el host objetivo\n");
      exit(0);
    }
    memset(tmp, 0, 1024);
    for (c = 0; c < strlen(target); c++) {
      if (target[c] == '/') break; 
      tmp[c] = target[c];
    }
    dir.s_addr = resuelve_a_ip(tmp);
    pmask = &target[c+1];
    mask = atoi(pmask);
    parse_mask(dir, mask);
    return;
  }

  dir.s_addr = resuelve_a_ip(target);
  new_tmp_host(dir);
}

void ping_hosts()
{
  int f = 0, g = 0, c, fd, d = 0, fds;
  struct host *ht, *ptr;
  unsigned short id = rand() % 0xffff;
  struct in_addr dir;
  int acks_env, pasadas = 0, pings_env, ant = 0, salir = 0, e = 0;
  struct in_addr mdir;

  fd = misocket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (fd < 0) {
    printf(" Error creando socket: %s\n", strerror(errno));
    fin(0);
  }

  c = fcntl(fd, F_GETFL);
  fcntl(fd, F_SETFL, c | O_NONBLOCK);

  printf("Enviando pings\n");
  fflush(stdout);

  if (verbose > 1) printf("Pings enviados: ");
  fflush(stdout);

  for (;;) {
      
    pings_env = 0;
    ht = tmp_hosts;
    f = 0;

    for (c = 0; c < ntmphosts; c++) {
      
      if (busca_host(ht->dir)) {
	ht = ht->next;
	continue;
      }

      host_timeout(ht);
      icmp_ping(ht->dir, id, fd);
      
      pings_env++;
      ht = ht->next;

      if ((f++) == MAX_PACKETS) {
	g = 0;
	set_timeout();
	while (g < f) {
	  dir = icmp_wait_pong(id, 1);
	  if (dir.s_addr) {
	    new_host(dir);
	    g++;
	  }
	  if (get_timeout()) break;
	}
	f = 0;
	g = 0;
      }
    }
    
    g = 0;
    set_timeout();
    while (g < f) {
      dir = icmp_wait_pong(id, 1);
      if (dir.s_addr) {
	new_host(dir);
	g++;
      }
      if (get_timeout()) break;
    }
    
    if (verbose > 1) printf("%i, ", pings_env);
    fflush(stdout);

    if (!pings_env) break;

    if (pings_env == ant) salir++;
    else salir = 0;
    ant = pings_env;
    if (salir == 2) break;
  }
  
  if (verbose > 1) printf("\n");

  if ((!solo_pings) && (pings_env)) {

    fds = misocket(AF_INET, SOCK_RAW, IPPROTO_RAW);
    if (fds < 0) {
      printf(" Error creando socket: %s\n", strerror(errno));
      fin(0);
    }
    
    d = fcntl(fds, F_GETFL);
    fcntl(fds, F_SETFL, d | O_NONBLOCK);
    
    mdir = mi_ip();
    
    printf("Enviando tcp acks\n");
    fflush(stdout);

    if (verbose > 1) printf("Acks enviados: ");
    fflush(stdout);
    
    pasadas = ant = salir = 0;
    
    for (;;) {
      
      acks_env = 0;
      ht = tmp_hosts;
      d = 0;
      
      for (c = 0; c < ntmphosts; c++) {
	
	if (busca_host(ht->dir)) {
	  ht = ht->next;
	  continue;
	}

	host_timeout(ht);
	TCP(fds, mdir, ht->dir, id, ack_port, TH_ACK);
     
	acks_env++;
	ht = ht->next;
	
	if ((d++) == MAX_PACKETS) {    
	  e = 0;
	  set_timeout();
	  while (e < d) {
	    dir = TCP_wait_ack(ack_port, 1);
	    if (dir.s_addr) {
	      new_host(dir);
	      e++;
	    }
	    if (get_timeout()) break;
	  }
	}
      }
      
      e = 0;
      set_timeout();
      while (e < d) {
	dir = TCP_wait_ack(ack_port, 1);
	if (dir.s_addr) {
	  new_host(dir);
	  e++;
	}
	if (get_timeout()) break;
      }

      if (verbose > 1) printf("%i, ", acks_env);
      fflush(stdout);

      if (!acks_env) break;

      if (acks_env == ant) salir++;
      else salir = 0;
      ant = acks_env;
      if (salir == 2) break;

    }
    close(fds);
    if (verbose > 1) printf("\n");  
  }

  close(fd);

  ht = tmp_hosts;
  while (ht) {
    ptr = ht->next;
    free(ht);
    ht = ptr;
  }
  tmp_hosts = NULL;

  if (verbose > 1) {
    printf("Respuestas: %i host%s activo%s\n", nhosts, (nhosts != 1) ? "s" : "", (nhosts != 1) ? "s" : "");
    fflush(stdout);
    
    if (logfile) fprintf(logfile, "Respuestas: %i host%s activo%s\n", nhosts, (nhosts > 1) ? "s" : "", (nhosts > 1) ? "s" : "");
  }

  resuelve_host();

  if (solohosts) {
    ht = lista_hosts;
    printf("\n");
    for (c = 0; c < nhosts; c++) {
      printf("%s (%s)\n", ht->name, inet_ntoa(ht->dir));
      if (logfile) fprintf(logfile, "%s (%s)\n", ht->name, inet_ntoa(ht->dir));
      ht = ht->next;
    }
  }
}

void ack_hosts()
{
  int e, f, c, d, fds;
  struct host *ht, *ptr;
  unsigned short id = rand() % 0xffff;
  struct in_addr dir, mdir = mi_ip();
  int acks_env, ant = 0, salir = 0;

  ht = tmp_hosts;

  fds = misocket(AF_INET, SOCK_RAW, IPPROTO_RAW);
  if (fds < 0) {
    printf(" Error creando socket: %s\n", strerror(errno));
    fin(0);
  }

  d = fcntl(fds, F_GETFL);
  fcntl(fds, F_SETFL, d | O_NONBLOCK);

  printf("Enviando tcp acks\n");
  fflush(stdout);

  if (verbose > 1) printf("Acks enviados: ");
  fflush(stdout);
  
  for (;;) {
  
    acks_env = 0;
    ht = tmp_hosts;
    d = 0;
    
    for (c = 0; c < ntmphosts; c++) {
      
      if (busca_host(ht->dir)) {
	ht = ht->next;
	continue;
      }
      
      host_timeout(ht);    
      TCP(fds, mdir, ht->dir, id, ack_port, TH_ACK);
     
      acks_env++;
      ht = ht->next;
      
      if ((d++) == MAX_PACKETS) {    
	e = 0;
	set_timeout();
	while (e < d) {
	  dir = TCP_wait_ack(ack_port, 1);
	  if (dir.s_addr) {
	    new_host(dir);
	    e++;
	  }
	  if (get_timeout()) break;
	}
      }
    }
    
    e = 0;
    f = 0;
    set_timeout();
    while (e < d) {
      dir = TCP_wait_ack(ack_port, 1);
      if (dir.s_addr) {
	new_host(dir);
	e++;
      }
      if (get_timeout()) break;
    }

    if (verbose > 1) printf("%i, ", acks_env);
    fflush(stdout);

    if (!acks_env) break;

    if (acks_env == ant) salir++;
    else salir = 0;
    ant = acks_env;
    if (salir == 2) break;

  }

  if (verbose > 1) printf("\n");

  close(fds);

  ht = tmp_hosts;
  while (ht) {
    ptr = ht->next;
    free(ht);
    ht = ptr;
  }
  tmp_hosts = NULL;

  if (verbose > 2) {
    printf("Respuestas: %i host%s activo%s\n", nhosts, (nhosts != 1) ? "s" : "", (nhosts != 1) ? "s" : "");
    fflush(stdout);
  
    if (logfile) fprintf(logfile, "Respuestas: %i host%s activo%s\n", nhosts, (nhosts > 1) ? "s" : "", (nhosts > 1) ? "s" : "");
  }

  resuelve_host();

  if (solohosts) {
    ht = lista_hosts;
    printf("\n");
    for (c = 0; c < nhosts; c++) {
      printf("%s (%s)\n", ht->name, inet_ntoa(ht->dir));
      if (logfile) fprintf(logfile, "%s (%s)\n", ht->name, inet_ntoa(ht->dir));
      ht = ht->next;
    }
  }
}
