
#include "netscan.h"

struct psdhdr {
  u_int32_t saddr;
  u_int32_t daddr;
  char zero;
  char prot;
  u_int16_t len;
};

extern struct in_addr ip_src;
extern void fin(int);
extern int sd;
extern int head_offset;
extern int misocket(int, int, int);

static unsigned short tcheck(struct tcphdr*, int, unsigned long, unsigned long);

extern struct host *tmp_hosts;
extern char *itz;

int read_ip_packet(char *dst, int len);

void TCP(int fd, struct in_addr sdir, struct in_addr ddir, 
	 unsigned short sp, unsigned short dp, unsigned char flags)
{
  char buffer[1024];
  struct ip *iph = (struct ip*) buffer;
  struct tcphdr *tcph = (struct tcphdr*) (buffer + sizeof(struct ip));
  struct sockaddr_in to;
    
  memset(buffer, 0, 1024);

  tcph->th_sport = htons(sp);
  tcph->th_dport = htons(dp);
  tcph->th_seq = rand();
  tcph->th_off = 5;
  tcph->th_flags = flags;
  tcph->th_sum = tcheck(tcph, sizeof(struct tcphdr), sdir.s_addr, ddir.s_addr);

  iph->ip_hl = 5;
  iph->ip_v = 4;
  iph->ip_len = htons(sizeof(struct ip) + sizeof(struct tcphdr));
  iph->ip_id = rand() % 0xffff;
  iph->ip_ttl = 0xff;
  iph->ip_p = 6;
  iph->ip_src.s_addr = sdir.s_addr;
  iph->ip_dst.s_addr = ddir.s_addr;

  to.sin_family = AF_INET;
  to.sin_port = htons(dp);
  to.sin_addr.s_addr = ddir.s_addr;

  if (sendto(fd, buffer, sizeof(struct ip) + sizeof(struct tcphdr), 0, 
	     (struct sockaddr_in*) &to, sizeof(to)) != (sizeof(struct ip) + sizeof(struct tcphdr))) { 
    printf(" Error: TCP(): sendto() devolvio -1\nsendto(): %s\n", strerror(errno));
    fin(0);
  }

}

void IP_TCP_OPT(int fd, struct in_addr sdir, struct in_addr ddir,
	    u_int8_t ttl, u_int32_t nseq, unsigned short sp, unsigned short dp,
		unsigned char flags, int lenopts, char *opts)
{
  char buffer[1024];
  struct ip *iph = (struct ip*) buffer;
  struct tcphdr *tcph = (struct tcphdr*) (buffer + sizeof(struct ip));
  struct sockaddr_in to;
    
  memset(buffer, 0, 1024);

  if (lenopts % 4) {
    printf("Error: IP_TCP_OPT(): Longitud de opciones ilegal\n");
    exit(0);
  }
  
  if (lenopts) memcpy(buffer + 40, opts, lenopts);

  tcph->th_sport = htons(sp);
  tcph->th_dport = htons(dp);
  tcph->th_seq = nseq;
  tcph->th_off = 5 + (lenopts / 4);
  tcph->th_flags = flags;
  tcph->th_sum = tcheck(tcph, sizeof(struct tcphdr) + lenopts, sdir.s_addr, ddir.s_addr);

  iph->ip_hl = 5;
  iph->ip_v = 4;
  iph->ip_len = htons(sizeof(struct ip) + sizeof(struct tcphdr) + lenopts);
  iph->ip_id = rand() % 0xffff;
  iph->ip_ttl = ttl;
  iph->ip_p = 6;
  iph->ip_src.s_addr = sdir.s_addr;
  iph->ip_dst.s_addr = ddir.s_addr;

  to.sin_family = AF_INET;
  to.sin_port = htons(dp);
  to.sin_addr.s_addr = ddir.s_addr;

  if (sendto(fd, buffer, sizeof(struct ip) + sizeof(struct tcphdr) + lenopts, 0, 
	     (struct sockaddr_in*) &to, sizeof(to)) != (sizeof(struct ip) + sizeof(struct tcphdr))) {
    if (errno == EAGAIN) return; 
    
    printf(" Error: IP_TCP_OPT(): sendto() devolvio -1\nsendto(): %s\n", strerror(errno));
    fin(0);
  }

}

struct in_addr TCP_wait_ack(unsigned short sp, int Timeout)
{
  char buffer[1024];
  struct ip *iph;
  struct tcphdr *tcph;
  struct in_addr ret;
  int timeout = Timeout;

  ret.s_addr = 0;

  for (;;) {

    if (!read_ip_packet(buffer, 20)) {
      timeout--;
      if (!timeout) return ret;
      usleep(1000000);
      continue;
    }

    iph = (struct ip*) buffer;

    if (iph->ip_p != 6) continue;

    if (!ip_src.s_addr) ip_src.s_addr = iph->ip_dst.s_addr;
    
    tcph = (struct tcphdr*) (buffer + (iph->ip_hl * 4));

    if (sp != htons(tcph->th_sport)) continue;

    ret.s_addr = iph->ip_src.s_addr;
    return ret;
  }
}

unsigned short tcheck( struct tcphdr *tcp, int len, unsigned long saddr, 
		       unsigned long daddr)
{
  char buff[4096];
  unsigned short *ptr = (unsigned short*) buff;
  struct psdhdr *psd = (struct psdhdr*) buff;
  u_int32_t ret = 0,c;
  char *pad;

  bzero(buff,4096);
  memcpy(&buff[sizeof(struct psdhdr)],tcp,len);

  psd->saddr = saddr;
  psd->daddr = daddr;
  psd->zero = 0;
  psd->prot = 6;
  psd->len = htons(len);
  
  for (c = 0; c < (len + sizeof(struct psdhdr)); c+= 2)  
    ret += (*(ptr++));
   
  if (c < (len + sizeof(struct psdhdr))) {
    ptr--;
    pad = (char*) ptr;
    pad++;
    ret += ((*pad) << 8);
    }
  ret = (ret >> 16) + (ret & 0xffff);
  return ~ret;
} 

int read_ip_packet(char *dst, int len)
{
  char buf[4096];
  struct ip *iph;
  int r;

  while (1) {

    memset(buf, 0, 4096);
    
    if ((r = read(sd, buf, 4096)) < 1) return 0; 
    
    iph = (struct ip*) (buf + head_offset);
    
    memcpy(dst, iph, len + 20);
    return 1;
  }
}

struct in_addr mi_ip()
{
  struct ifreq *ifr;
  struct ifconf ifc;
  int fd = misocket(AF_INET, SOCK_DGRAM, 0);
  int c, d = 1;
  
  if (!ip_src.s_addr) {
   
    bzero(&ifc, sizeof(ifc));
    if (ioctl(fd, SIOCGIFCONF, &ifc) < 0) {
      printf(" Error: no puedo saber tu IP. Prueba con -s <ip>\n");
      fin(0);
    }

    ifr = (struct ifreq*) malloc(ifc.ifc_len);
    ifc.ifc_buf = (char*) ifr;
    
    if (ioctl(fd, SIOCGIFCONF, &ifc) < 0) {
      printf(" Error: no puedo saber tu IP. Prueba con -s <ip>\n");
      fin(0);
    }

    c = (ifc.ifc_len / sizeof(struct ifreq));

    for (d = 0; d < c; d++)        
      if (!strncmp(ifr[d].ifr_name, itz, 2)) {
	ioctl(fd, SIOCGIFFLAGS, &ifr[d]);
	if ((ifr[d].ifr_flags & IFF_UP) == IFF_UP)
	  return (*(struct sockaddr_in*)&ifr[d].ifr_addr).sin_addr;
      }
    printf(" Error: no puedo saber tu IP. Prueba con -s <ip>\n");
    fin(0);
  }
  
  return ip_src;
}

char *get_ifaz()
{
  FILE *f;
  char device[256];
  u_int32_t dest;
  char gtw[256];
  int flags;
  int refcnt;
  int use;
  int metric;
  int mtu, window, irtt;
  u_int32_t mask;
  char ch;
  char *default_dev = NULL;
  
  f = fopen("/proc/net/route", "r");
 
  if (!f) {
    perror("open():");
    exit(0);
  }
  
  while (!feof(f)) {
    
    for (;;) {
      fread(&ch, 1, 1, f);
      if (ch == '\n') break;
      if (feof(f)) {
	printf("Error: no se en que interfaz escuchar paquetes. Prueba con -i\n");
	exit(0);
      }
    }
    
    fscanf(f, "%s\t%x\t%s\t%x\t%i\t%i\t%i\t%x\t%x\t%x\t%x",
	   device, &dest, gtw, &flags, &refcnt, &use, &metric, &mask, &mtu, &window, &irtt);

    if (!(flags & 0x1)) continue;

    if ((tmp_hosts->dir.s_addr & mask) == (dest & mask)) {
      fclose(f);
      if (default_dev) free(default_dev);
      default_dev = (char*) malloc(strlen(device) + 1);
      strncpy(default_dev, device, strlen(device));
      return default_dev;
    }
    
    if (!dest) {
      default_dev = (char*) malloc(strlen(device) + 1);
      strncpy(default_dev, device, strlen(device));
    }
  }

  fclose(f);

  if (default_dev) return default_dev;

  printf("Error: no se en que interfaz escuchar paquetes. Prueba con -i\n");
  exit(0);

}

