
#include "netscan.h"

struct host *lista_makinas;

int nmak = 0;
int nips = 0;

extern struct host *lista_hosts;
extern struct host *tmp_hosts;

extern int ntmphosts;
extern int nhosts;
extern FILE *logfile;
extern int solohosts;
extern int npuertos;
extern int MAX_PACKETS;

extern void resuelve_host();
extern struct host *busca_host(struct in_addr);
extern void icmp_tstamp(struct in_addr, unsigned short, int);
extern struct in_addr icmp_wait_tstamp(unsigned short, int, u_int32_t*);
extern struct host *get_tmp_host(struct in_addr );
extern void set_timeout();
extern int get_timeout();
extern void host_timeout(struct host*);
extern void calcula_host_timeout(struct host *ht);
extern int misocket(int, int, int);
extern int scan_timeout_sec, scan_timeout_msec;

void computa_tiempos();

static void new_host_dif(struct in_addr dir, u_int32_t dif)
{
  struct host *tmp, *ptr = lista_hosts, *new;

  if (busca_host(dir)) return;

  tmp = get_tmp_host(dir);  

  if (!ptr) {
    lista_hosts = (struct host*) malloc(sizeof(struct host));
    memset(lista_hosts, 0, sizeof(struct host));
    new = lista_hosts;
    new->dir.s_addr = dir.s_addr;
    new->dif = dif;
    new->puertos = (struct puerto*) malloc(npuertos * sizeof(struct puerto));
    if (tmp) {
      new->seq = tmp->seq;
      memcpy(&new->time1, &tmp->time1, sizeof(tmp->time1));
      calcula_host_timeout(new);
    } else {
      new->timeout.tv_sec = scan_timeout_sec;
      new->timeout.tv_usec = scan_timeout_msec;
    }
    nhosts = 1;
    return;
  }

  while (ptr->next) ptr = ptr->next;
  new = (struct host*) malloc(sizeof(struct host));
  memset(new, 0, sizeof(struct host));
  new->dir.s_addr = dir.s_addr;
  new->dif = dif;
  new->puertos = (struct puerto*) malloc(npuertos * sizeof(struct puerto));
  if (tmp) {
    new->seq = tmp->seq;
    memcpy(&new->time1, &tmp->time1, sizeof(tmp->time1));
    calcula_host_timeout(new);
  } else {
    new->timeout.tv_sec = scan_timeout_sec;
    new->timeout.tv_usec = scan_timeout_msec;
  }
  ptr->next = new;
  nhosts++;
}

void tstamp_hosts()
{
  int f = 0, g = 0, c, fd, d = 0;
  struct host *ht, *ptr, *mak, *mak2;
  unsigned short id = rand() % 0xffff;
  struct in_addr dir;
  u_int32_t dif;
  int tstamp_env, ant = 0, salir = 0;

  ht = tmp_hosts;

  fd = misocket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
  if (fd < 0) {
    printf(" Error creando socket: %s\n", strerror(errno));
    exit(0);
  }

  c = fcntl(fd, F_GETFL);
  fcntl(fd, F_SETFL, c | O_NONBLOCK);

  printf("Enviando TimeStamp Requests\n");
  fflush(stdout);

  if (verbose > 1) printf("Tstamps enviadas: ");
  fflush(stdout);

  for (;;) {

    tstamp_env = 0;
    ht = tmp_hosts;
    f = 0;
    
    for (c = 0; c < ntmphosts; c++) {

      if (busca_host(ht->dir)) {
	ht = ht->next;
	continue;
      }

      host_timeout(ht);
      icmp_tstamp(ht->dir, id, fd);
      
      tstamp_env++;
      ht = ht->next;

      if ((f++) == MAX_PACKETS) {
	g = 0;
	set_timeout();
	while (g < f) {
	  dir = icmp_wait_tstamp(id, 1, &dif);
	  if (dir.s_addr) {
	    new_host_dif(dir, dif);
	    g++;
	  }
	  if (get_timeout()) break;
	}
	f = 0;
	g = 0;
      }
    }
    
    g = 0;
    d = 0;
    set_timeout();
    while (g < f) {
      dir = icmp_wait_tstamp(id, 1, &dif);
      if (dir.s_addr) {
	new_host_dif(dir, dif);
	g++;
      }
      if (get_timeout()) break;
    }
    
    if (verbose > 1) printf("%i, ", tstamp_env);
    fflush(stdout);

    if (!tstamp_env) break;

    if (tstamp_env == ant) salir++;
    else salir = 0;
    ant = tstamp_env;
    if (salir == 2) break;

    
  }

  if (verbose > 1) printf("\n");

  close(fd);

  ht = tmp_hosts;
  while (ht) {
    ptr = ht->next;
    free(ht);
    ht = ptr;
  }
  tmp_hosts = NULL;
 
  if (verbose > 1) {
    printf("Respuestas: %i host%s activo%s\n", nhosts, (nhosts != 1) ? "s" : "", (nhosts != 1) ? "s" : "");
    fflush(stdout);
    
    if (logfile) fprintf(logfile, "Respuestas: %i host%s activo%s\n", nhosts, (nhosts > 1) ? "s" : "", (nhosts > 1) ? "s" : "");
  }

  resuelve_host();

  computa_tiempos();

  if (solohosts) {
    ht = lista_hosts;
    printf("\n");
    for (c = 0; c < nhosts; c++) {
      mak = ht->next_mak;
      mak2 = mak;
      printf("%s (%s)\n", ht->name, inet_ntoa(ht->dir));
      while (mak) {
	printf("\t|----> %s (%s)\n", mak->name, inet_ntoa(mak->dir));
	mak = mak->next_mak;
      }
      if (logfile) {
	fprintf(logfile, "%s (%s)\n", ht->name, inet_ntoa(ht->dir));
	while (mak2) {
	  fprintf(logfile, "\t|----> %s (%s)\n", mak2->name, inet_ntoa(mak2->dir));
	  mak2 = mak2->next_mak;
	}
      }
      ht = ht->next;
    }
  }
}

struct host *busca_mak(u_int32_t dif)
{
  struct host *ptr = lista_makinas;
  
  while (ptr) {
    if (ptr->dif == dif) return ptr;
    ptr = ptr->next;
  }
  return NULL;
}

void nueva_makina_misma_ip(struct host *mak, struct host *new)
{
  mak->nips++;
  while (mak->next_mak) mak = mak->next_mak;
  mak->next_mak = new;
  nips++;
}

void nueva_makina(struct host *new2)
{
  struct host *ptr = lista_makinas, *new;

  nips++;
  nmak++;

  if (!ptr) {
    lista_makinas = (struct host*) malloc(sizeof(struct host));
    memset(lista_makinas, 0, sizeof(struct host));
    new = lista_makinas;
    new->dir.s_addr = new2->dir.s_addr;
    strcpy(new->name, new2->name);
    new->dif = new2->dif;
    new->puertos = new2->puertos;
    new->nips = 1;
    memcpy(&new->timeout, &new2->timeout, sizeof(new->timeout));
    return;
  }

  while (ptr->next) ptr = ptr->next;
  new = (struct host*) malloc(sizeof(struct host));
  memset(new, 0, sizeof(struct host));
  new->dir.s_addr = new2->dir.s_addr;
  new->dif = new2->dif;
  ptr->next = new;
  new->puertos = new2->puertos;
  new->nips = 1;
  memcpy(&new->timeout, &new2->timeout, sizeof(new->timeout));
  strcpy(new->name, new2->name);
}

void computa_tiempos()
{
  struct host *mak, *ptr = lista_hosts;

  if (verbose) printf("Buscando hosts con varias IPs:");
  fflush(stdout);

  while (ptr) {

    if ((mak = busca_mak(ptr->dif))) {
      nueva_makina_misma_ip(mak, ptr);
    } else nueva_makina(ptr);
    
    ptr = ptr->next;
  }

  lista_hosts = lista_makinas;
  nhosts = nmak;

  if (verbose) {
    if (nmak > 1) printf(" %i hosts diferentes en %i IPs\n", nmak, nips);
    else printf(" 1 host en %i IPs\n", nips);
    fflush(stdout);
  }
}

