
#include "netsniff.h"

struct paquete {
  struct ip iph;
  struct tcphdr tcph;
  char datos[4096];
  int len;
  struct paquete *next;
};

struct conex {
  u_int32_t sa, da;
  u_int16_t sp, dp;
  int timeout;
  int len;
  u_int32_t inicio;
  char ns[1024];
  char nd[1024];
  struct paquete *paquetes;
  struct conex *next;
  struct conex *prev;
};

struct conex *lista_conex = NULL;

int nconex = 0;

extern u_int32_t swaps(u_int32_t );

extern char *logfile;

extern int multiple_logs;

int log_count = 0;

void add_packet(struct ip*, struct tcphdr*);
void fin_conex(struct ip*, struct tcphdr*);
void addr_to_ascii(char *, u_int32_t , int);

extern int mcons;
extern int mdatos;
extern int nodatos;

extern int con_timeout; 

void mira_timeout()
{
  struct conex *tmp = lista_conex;
  struct ip iph;
  struct tcphdr tcph;
  
  while (tmp) {
    
    tmp->timeout++;
    
    if (tmp->timeout >= con_timeout) {
      iph.ip_src.s_addr = tmp->sa;
      iph.ip_dst.s_addr = tmp->da;

      tcph.th_sport = tmp->sp;
      tcph.th_dport = tmp->dp;

      fin_conex(&iph, &tcph);

      tmp = lista_conex;
      continue;
    }
    tmp = tmp->next;
  }
}


struct conex *busca_conex(struct ip *iph, struct tcphdr *tcph)
{
  struct conex *tmp = lista_conex; 

  while (tmp) {
    if (tmp->sa == iph->ip_src.s_addr)
      if (tmp->da == iph->ip_dst.s_addr)
	if (tmp->sp == tcph->th_sport)
	  if (tmp->dp == tcph->th_dport) return tmp;
    tmp = tmp->next;
  }

  tmp = lista_conex;
  while (tmp) {
    if (tmp->da == iph->ip_src.s_addr)
      if (tmp->sa == iph->ip_dst.s_addr)
	if (tmp->dp == tcph->th_sport)
	  if (tmp->sp == tcph->th_dport) return tmp;
    tmp = tmp->next;
  }
  return NULL;
}

extern int nports;
extern u_int16_t ports[255];
extern int all_ports;

int is_listen(u_int16_t p)
{
  int c = 0;
  while (ports[c]) {
    if (ports[c] == p) return 1;
    c++;
  }
  return 0;
}

void new_conex(struct ip *iph, struct tcphdr *tcph)
{
  struct conex *tmp = lista_conex, *new;
  struct hostent *ht;

  if (!all_ports)
    if (!is_listen(swaps(tcph->th_dport))) return;

  if (mcons)
    if ((nconex + 1) > mcons) return;

  new = (struct conex*) malloc(sizeof(struct conex));
  memset(new, 0, sizeof(struct conex));

  new->sa = iph->ip_src.s_addr;
  new->da = iph->ip_dst.s_addr;
  new->sp = tcph->th_sport;
  new->dp = tcph->th_dport;
  new->inicio = time(NULL);

  ht = gethostbyaddr((char*)&iph->ip_src, 4, AF_INET);
  if (ht) strncpy(new->ns, ht->h_name, 1024);
  ht = gethostbyaddr((char*)&iph->ip_dst, 4, AF_INET);
  if (ht) strncpy(new->nd, ht->h_name, 1024);

  nconex++;

  if (!tmp) {
    lista_conex = new;
    return;
  }

  while (tmp->next) tmp = tmp->next;

  tmp->next = new;
  new->prev = tmp;
}

void procesa(struct ip *iph)
{
  struct tcphdr *tcph;

  tcph = (struct tcphdr*) ((char*)iph + (iph->ip_hl * 4));
  
  if (tcph->th_flags == TH_SYN) { 
    if (!busca_conex(iph, tcph)) new_conex(iph, tcph);
    return;
  }
  
  if ((tcph->th_flags & TH_RST) || (tcph->th_flags & TH_FIN)) {
    if (busca_conex(iph, tcph)) fin_conex(iph, tcph);
    return;
  }
  
  add_packet(iph, tcph);
}


void add_packet(struct ip *iph, struct tcphdr *tcph)
{
  struct conex *tmp;
  struct paquete *pn, *pkt;
  int len;
  char *datos;

  tmp = busca_conex(iph, tcph);

  if (!tmp) return;

  pkt = (struct paquete*) malloc(sizeof(struct paquete));  
  memset(pkt, 0, sizeof(struct paquete));

  memcpy(&pkt->iph, iph, sizeof(struct ip));
  memcpy(&pkt->tcph, tcph, sizeof(struct tcphdr));

  datos = (char*) ((char*) tcph + (tcph->th_off * 4));
  len = swaps(iph->ip_len) - (iph->ip_hl * 4) - (tcph->th_off * 4);
  pkt->len = len;
  memcpy(pkt->datos, datos, len);

  if (mdatos)
    if ((tmp->len + len) > mdatos) {
      free(pkt);
      return;
    }
  
  pn = tmp->paquetes;

  tmp->len += len;

  if (!pn) {
    tmp->paquetes = pkt;
    return;
  }

  while (pn->next) {
    if (pn->iph.ip_id == iph->ip_id) {
      free(pkt);
      return;
    }
    pn = pn->next;
  }

  if (pn->iph.ip_id == iph->ip_id) {
    free(pkt);
    return;
  }

  pn->next = pkt;
}

char retdur[128];

void duracion(u_int32_t sec)
{
  u_int32_t seg, min, hor;
  hor = sec / 3600;
  min = (sec % 3600) / 60;
  seg = (sec % 3600) % 60;
  memset(retdur, 0, 128);
  sprintf(retdur, "%i:%02i:%02i", (int) hor, (int) min, (int) seg);
}

void fin_conex(struct ip *iph, struct tcphdr *tcph)
{
  struct paquete *pkt, *pn;
  struct conex *tmp;
  FILE *log;
  char buff[128], buff2[1024];
  struct in_addr dir;
  u_int32_t t1;
  struct servent *pe;

  tmp = busca_conex(iph, tcph);
  if (!tmp) return;

  nconex--;

  if (!multiple_logs) log = fopen(logfile, "a");  
  else {
    sprintf(buff2, "%s%i", logfile, log_count++);
    log = fopen(buff2, "a");
  }

  if ((log) && (!nodatos ? 1 : tmp->len)) {
   
    fprintf(log,"\n");
    fprintf(log,"---------------------------------------------------------------------------\n");
	
    memset(buff, 0, 128); 
    addr_to_ascii(buff, tmp->sa, 128);
    dir.s_addr = tmp->sa;
    fprintf(log," Origen   : %s (%s) [%i]\n", buff, inet_ntoa(dir), (int) swaps(tmp->sp));

    memset(buff, 0, 128); 
    addr_to_ascii(buff, tmp->da, 128);
    dir.s_addr = tmp->da;
    fprintf(log," Destino  : %s (%s) [%i]\n", buff, inet_ntoa(dir), (int) swaps(tmp->dp));
    
    pe = getservbyport(tmp->dp, "tcp");
    if (pe) fprintf(log," Sesion   : %s\n", pe->s_name);
    else fprintf(log, " Sesion   : desconocida\n");

    t1 = time(NULL);
    duracion((t1 - tmp->inicio));
    fprintf(log," Inicio   : %s Duracion : %s secs.\n",ctime(&tmp->inicio), retdur);
    fprintf(log," Datos    : %i bytes Conexiones activas: %i\n",tmp->len, nconex);

    fprintf(log,"---------------------------------------------------------------------------");
    fprintf(log,"\n");

    pkt = tmp->paquetes;
    while (pkt) {
      fwrite(pkt->datos, 1, pkt->len, log);
      pn = pkt;
      pkt = pkt->next;
      free(pn);
      }
    fprintf(log,"\n");
  }
  fclose(log);

  if (tmp == lista_conex) {
    if (tmp->next) {
      lista_conex = tmp->next;
      free(tmp);
      return;
    }
    free(tmp);
    lista_conex = NULL;
    return;
  }

  if ((tmp->next) && (tmp->prev)) {
    tmp->next->prev = tmp->prev;
    tmp->prev->next = tmp->next;
    free(tmp);
    return;
  }

  if ((tmp->next) && (!tmp->prev)) {
    tmp->next->prev = tmp->prev;
    free(tmp);
    return;
  }

  if ((!tmp->next) && (tmp->prev)) {
    tmp->prev->next = tmp->next;
    free(tmp);
    return;
  }

}

void addr_to_ascii(char *nombre, u_int32_t sa, int max)
{
  struct hostent *he;
  struct in_addr dir;
  dir.s_addr = sa;

  he = gethostbyaddr( (char*) &dir, sizeof(dir), AF_INET);
  if (!he) return;
  if (!he->h_name) return;
  strncpy(nombre, he->h_name, max);
}
